
;;--------------------------------------------------------------------
;; s8vector basic API
;;
(define-cproc s8vector? (obj) ::<boolean> SCM_S8VECTORP)

;; we need this in string, for the UNBOX macro is substituted to the
;; literal C code instead of CISE code.
"#define s8unboxer(filler, fill) filler = (signed char)Scm_GetInteger8Clamp(fill, SCM_CLAMP_ERROR, NULL)"

(define-cproc make-s8vector (length::<fixnum> :optional (fill 0))
  (let* ([filler :: (signed char)])
    (s8unboxer filler fill)
    (return (Scm_MakeS8Vector length filler))))

(define-cproc s8vector (:optarray (elts nelts 10) :rest args)
  :fast-flonum
  (cond [(SCM_NULLP args)
         (return (Scm_ObjArrayToS8Vector elts nelts SCM_CLAMP_ERROR))]
        [else
         (let* ([i::int (- nelts 1)] [p args])
           (for [() (>= i 0) (post-- i)]
                (SCM_FLONUM_ENSURE_MEM (aref elts i))
                (set! p (Scm_Cons (aref elts i) p)))
           (return (Scm_ListToUVector SCM_CLASS_S8VECTOR p SCM_CLAMP_ERROR)))]))

(define-cproc s8vector-length (v::<s8vector>) ::<int> SCM_S8VECTOR_SIZE)

;; This is only called via apply.  The standard call will be inlined
;; as UVEC-REF instruction.
(define-cproc s8vector-ref
  (v::<s8vector> i::<fixnum> :optional fallback) :fast-flonum
  (setter s8vector-set!)
  (return (Scm_VMUVectorRef v (Scm_UVectorType (Scm_ClassOf (SCM_OBJ v)))
                            i fallback)))

(define-cproc s8vector-set!
  (v::<s8vector> i::<fixnum> val :optional clamp) :fast-flonum
  (return (Scm_S8VectorSet v i val (clamp_arg clamp))))

(define-cproc s8vector-copy
  (v::<s8vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_S8VectorCopy)

;; NB: this function accomodates to two APIs.
;; The 'modern' API, compatible to srfi-13 and srfi-43, takes arguments:
;;    (dst dstart src :optional sstart send)
;; The old API only takes:
;;    (dst src)
;; We dispatch by the second argument.  The old API is deprecated, but
;; kept for the existing code.
(define-cproc s8vector-copy!
  (dst::<s8vector> dstart :optional src (sstart::<int> 0) (send::<int> -1))
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (cond
   [(SCM_INTEGERP dstart) ; new API
    (unless (SCM_S8VECTORP src)
      (Scm_Error "s8vector-copy!: argument is not a matching uvector: %S"src))
    (return (Scm_S8VectorCopyX dst (Scm_GetInteger dstart)
                  (SCM_S8VECTOR src) sstart send))]
   [(SCM_S8VECTORP dstart) ; old API
    (return (Scm_S8VectorCopyX dst 0 (SCM_S8VECTOR dstart) 0 -1))]
   [else (SCM_TYPE_ERROR dstart "integer") (return SCM_UNDEFINED)]))

;; multi-copy! - copy src repeatedly to different parts of dst.
(define-cproc s8vector-multi-copy!
  (dst::<s8vector> dstart::<fixnum> dstride::<fixnum>
   src::<s8vector> :optional (sstart::<fixnum> 0) (ssize::<fixnum> 0)
                               (sstride::<fixnum> -1) (count::<fixnum> -1))
  ::<void>
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (when (< sstride 0)
    (if (<= ssize 0) (set! sstride 0) (set! sstride ssize)))
  (when (<= ssize 0)
    (set! ssize (- (SCM_UVECTOR_SIZE src) sstart)))
  (let* ([di::long dstart] [si::long sstart]
         [dlen::long (SCM_UVECTOR_SIZE dst)]
         [slen::long (SCM_UVECTOR_SIZE src)]
         [c::long 0] [k::long]
         [cnt::long (?: (< count 0) LONG_MAX count)])
    (if (== ssize 1)
      ;; shortcut for optimization
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (set! (SCM_S8VECTOR_ELEMENT dst di)
                 (SCM_S8VECTOR_ELEMENT src si))
           (+= di dstride)
           (+= si sstride))
      ;; general case
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (let* ([nitems::long ssize])
             (when (< dlen (+ di nitems)) (set! nitems (- dlen di)))
             (when (< slen (+ si nitems)) (set! nitems (- slen si)))
             (memmove (+ (SCM_S8VECTOR_ELEMENTS dst) di)
                      (+ (SCM_S8VECTOR_ELEMENTS src) si)
                      (* nitems (sizeof (.type signed char)))))
           (+= di dstride)
           (+= si sstride)))))

(define-cproc s8vector->list
  (v::<s8vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_S8VectorToList)

(define-cproc list->s8vector (l::<list> :optional clamp)
  (return (Scm_ListToUVector SCM_CLASS_S8VECTOR l (clamp-arg clamp))))

(define-cproc s8vector-fill!
  (v::<s8vector> val :optional (start::<fixnum> 0) (end::<fixnum> -1))
  (let* ([filler :: (signed char)])
    (s8unboxer filler val)
    (return (Scm_S8VectorFill v filler start end))))

(define-cproc s8vector->vector
  (v::<s8vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_S8VectorToVector)

(define-cproc vector->s8vector
  (v::<vector> :optional (start::<fixnum> 0) (end::<fixnum> -1) clamp)
  (return (Scm_VectorToS8Vector v start end (clamp-arg clamp))))


;;--------------------------------------------------------------------
;; u8vector basic API
;;
(define-cproc u8vector? (obj) ::<boolean> SCM_U8VECTORP)

;; we need this in string, for the UNBOX macro is substituted to the
;; literal C code instead of CISE code.
"#define u8unboxer(filler, fill) filler = (u_char)Scm_GetIntegerU8Clamp(fill, SCM_CLAMP_ERROR, NULL)"

(define-cproc make-u8vector (length::<fixnum> :optional (fill 0))
  (let* ([filler :: (unsigned char)])
    (u8unboxer filler fill)
    (return (Scm_MakeU8Vector length filler))))

(define-cproc u8vector (:optarray (elts nelts 10) :rest args)
  :fast-flonum
  (cond [(SCM_NULLP args)
         (return (Scm_ObjArrayToU8Vector elts nelts SCM_CLAMP_ERROR))]
        [else
         (let* ([i::int (- nelts 1)] [p args])
           (for [() (>= i 0) (post-- i)]
                (SCM_FLONUM_ENSURE_MEM (aref elts i))
                (set! p (Scm_Cons (aref elts i) p)))
           (return (Scm_ListToUVector SCM_CLASS_U8VECTOR p SCM_CLAMP_ERROR)))]))

(define-cproc u8vector-length (v::<u8vector>) ::<int> SCM_U8VECTOR_SIZE)

;; This is only called via apply.  The standard call will be inlined
;; as UVEC-REF instruction.
(define-cproc u8vector-ref
  (v::<u8vector> i::<fixnum> :optional fallback) :fast-flonum
  (setter u8vector-set!)
  (return (Scm_VMUVectorRef v (Scm_UVectorType (Scm_ClassOf (SCM_OBJ v)))
                            i fallback)))

(define-cproc u8vector-set!
  (v::<u8vector> i::<fixnum> val :optional clamp) :fast-flonum
  (return (Scm_U8VectorSet v i val (clamp_arg clamp))))

(define-cproc u8vector-copy
  (v::<u8vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_U8VectorCopy)

;; NB: this function accomodates to two APIs.
;; The 'modern' API, compatible to srfi-13 and srfi-43, takes arguments:
;;    (dst dstart src :optional sstart send)
;; The old API only takes:
;;    (dst src)
;; We dispatch by the second argument.  The old API is deprecated, but
;; kept for the existing code.
(define-cproc u8vector-copy!
  (dst::<u8vector> dstart :optional src (sstart::<int> 0) (send::<int> -1))
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (cond
   [(SCM_INTEGERP dstart) ; new API
    (unless (SCM_U8VECTORP src)
      (Scm_Error "u8vector-copy!: argument is not a matching uvector: %S"src))
    (return (Scm_U8VectorCopyX dst (Scm_GetInteger dstart)
                  (SCM_U8VECTOR src) sstart send))]
   [(SCM_U8VECTORP dstart) ; old API
    (return (Scm_U8VectorCopyX dst 0 (SCM_U8VECTOR dstart) 0 -1))]
   [else (SCM_TYPE_ERROR dstart "integer") (return SCM_UNDEFINED)]))

;; multi-copy! - copy src repeatedly to different parts of dst.
(define-cproc u8vector-multi-copy!
  (dst::<u8vector> dstart::<fixnum> dstride::<fixnum>
   src::<u8vector> :optional (sstart::<fixnum> 0) (ssize::<fixnum> 0)
                               (sstride::<fixnum> -1) (count::<fixnum> -1))
  ::<void>
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (when (< sstride 0)
    (if (<= ssize 0) (set! sstride 0) (set! sstride ssize)))
  (when (<= ssize 0)
    (set! ssize (- (SCM_UVECTOR_SIZE src) sstart)))
  (let* ([di::long dstart] [si::long sstart]
         [dlen::long (SCM_UVECTOR_SIZE dst)]
         [slen::long (SCM_UVECTOR_SIZE src)]
         [c::long 0] [k::long]
         [cnt::long (?: (< count 0) LONG_MAX count)])
    (if (== ssize 1)
      ;; shortcut for optimization
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (set! (SCM_U8VECTOR_ELEMENT dst di)
                 (SCM_U8VECTOR_ELEMENT src si))
           (+= di dstride)
           (+= si sstride))
      ;; general case
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (let* ([nitems::long ssize])
             (when (< dlen (+ di nitems)) (set! nitems (- dlen di)))
             (when (< slen (+ si nitems)) (set! nitems (- slen si)))
             (memmove (+ (SCM_U8VECTOR_ELEMENTS dst) di)
                      (+ (SCM_U8VECTOR_ELEMENTS src) si)
                      (* nitems (sizeof (.type unsigned char)))))
           (+= di dstride)
           (+= si sstride)))))

(define-cproc u8vector->list
  (v::<u8vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_U8VectorToList)

(define-cproc list->u8vector (l::<list> :optional clamp)
  (return (Scm_ListToUVector SCM_CLASS_U8VECTOR l (clamp-arg clamp))))

(define-cproc u8vector-fill!
  (v::<u8vector> val :optional (start::<fixnum> 0) (end::<fixnum> -1))
  (let* ([filler :: (unsigned char)])
    (u8unboxer filler val)
    (return (Scm_U8VectorFill v filler start end))))

(define-cproc u8vector->vector
  (v::<u8vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_U8VectorToVector)

(define-cproc vector->u8vector
  (v::<vector> :optional (start::<fixnum> 0) (end::<fixnum> -1) clamp)
  (return (Scm_VectorToU8Vector v start end (clamp-arg clamp))))


;;--------------------------------------------------------------------
;; s16vector basic API
;;
(define-cproc s16vector? (obj) ::<boolean> SCM_S16VECTORP)

;; we need this in string, for the UNBOX macro is substituted to the
;; literal C code instead of CISE code.
"#define s16unboxer(filler, fill) filler = (short)Scm_GetInteger16Clamp(fill, SCM_CLAMP_ERROR, NULL)"

(define-cproc make-s16vector (length::<fixnum> :optional (fill 0))
  (let* ([filler :: (short)])
    (s16unboxer filler fill)
    (return (Scm_MakeS16Vector length filler))))

(define-cproc s16vector (:optarray (elts nelts 10) :rest args)
  :fast-flonum
  (cond [(SCM_NULLP args)
         (return (Scm_ObjArrayToS16Vector elts nelts SCM_CLAMP_ERROR))]
        [else
         (let* ([i::int (- nelts 1)] [p args])
           (for [() (>= i 0) (post-- i)]
                (SCM_FLONUM_ENSURE_MEM (aref elts i))
                (set! p (Scm_Cons (aref elts i) p)))
           (return (Scm_ListToUVector SCM_CLASS_S16VECTOR p SCM_CLAMP_ERROR)))]))

(define-cproc s16vector-length (v::<s16vector>) ::<int> SCM_S16VECTOR_SIZE)

;; This is only called via apply.  The standard call will be inlined
;; as UVEC-REF instruction.
(define-cproc s16vector-ref
  (v::<s16vector> i::<fixnum> :optional fallback) :fast-flonum
  (setter s16vector-set!)
  (return (Scm_VMUVectorRef v (Scm_UVectorType (Scm_ClassOf (SCM_OBJ v)))
                            i fallback)))

(define-cproc s16vector-set!
  (v::<s16vector> i::<fixnum> val :optional clamp) :fast-flonum
  (return (Scm_S16VectorSet v i val (clamp_arg clamp))))

(define-cproc s16vector-copy
  (v::<s16vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_S16VectorCopy)

;; NB: this function accomodates to two APIs.
;; The 'modern' API, compatible to srfi-13 and srfi-43, takes arguments:
;;    (dst dstart src :optional sstart send)
;; The old API only takes:
;;    (dst src)
;; We dispatch by the second argument.  The old API is deprecated, but
;; kept for the existing code.
(define-cproc s16vector-copy!
  (dst::<s16vector> dstart :optional src (sstart::<int> 0) (send::<int> -1))
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (cond
   [(SCM_INTEGERP dstart) ; new API
    (unless (SCM_S16VECTORP src)
      (Scm_Error "s16vector-copy!: argument is not a matching uvector: %S"src))
    (return (Scm_S16VectorCopyX dst (Scm_GetInteger dstart)
                  (SCM_S16VECTOR src) sstart send))]
   [(SCM_S16VECTORP dstart) ; old API
    (return (Scm_S16VectorCopyX dst 0 (SCM_S16VECTOR dstart) 0 -1))]
   [else (SCM_TYPE_ERROR dstart "integer") (return SCM_UNDEFINED)]))

;; multi-copy! - copy src repeatedly to different parts of dst.
(define-cproc s16vector-multi-copy!
  (dst::<s16vector> dstart::<fixnum> dstride::<fixnum>
   src::<s16vector> :optional (sstart::<fixnum> 0) (ssize::<fixnum> 0)
                               (sstride::<fixnum> -1) (count::<fixnum> -1))
  ::<void>
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (when (< sstride 0)
    (if (<= ssize 0) (set! sstride 0) (set! sstride ssize)))
  (when (<= ssize 0)
    (set! ssize (- (SCM_UVECTOR_SIZE src) sstart)))
  (let* ([di::long dstart] [si::long sstart]
         [dlen::long (SCM_UVECTOR_SIZE dst)]
         [slen::long (SCM_UVECTOR_SIZE src)]
         [c::long 0] [k::long]
         [cnt::long (?: (< count 0) LONG_MAX count)])
    (if (== ssize 1)
      ;; shortcut for optimization
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (set! (SCM_S16VECTOR_ELEMENT dst di)
                 (SCM_S16VECTOR_ELEMENT src si))
           (+= di dstride)
           (+= si sstride))
      ;; general case
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (let* ([nitems::long ssize])
             (when (< dlen (+ di nitems)) (set! nitems (- dlen di)))
             (when (< slen (+ si nitems)) (set! nitems (- slen si)))
             (memmove (+ (SCM_S16VECTOR_ELEMENTS dst) di)
                      (+ (SCM_S16VECTOR_ELEMENTS src) si)
                      (* nitems (sizeof (.type short)))))
           (+= di dstride)
           (+= si sstride)))))

(define-cproc s16vector->list
  (v::<s16vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_S16VectorToList)

(define-cproc list->s16vector (l::<list> :optional clamp)
  (return (Scm_ListToUVector SCM_CLASS_S16VECTOR l (clamp-arg clamp))))

(define-cproc s16vector-fill!
  (v::<s16vector> val :optional (start::<fixnum> 0) (end::<fixnum> -1))
  (let* ([filler :: (short)])
    (s16unboxer filler val)
    (return (Scm_S16VectorFill v filler start end))))

(define-cproc s16vector->vector
  (v::<s16vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_S16VectorToVector)

(define-cproc vector->s16vector
  (v::<vector> :optional (start::<fixnum> 0) (end::<fixnum> -1) clamp)
  (return (Scm_VectorToS16Vector v start end (clamp-arg clamp))))


;;--------------------------------------------------------------------
;; u16vector basic API
;;
(define-cproc u16vector? (obj) ::<boolean> SCM_U16VECTORP)

;; we need this in string, for the UNBOX macro is substituted to the
;; literal C code instead of CISE code.
"#define u16unboxer(filler, fill) filler = (u_short)Scm_GetIntegerU16Clamp(fill, SCM_CLAMP_ERROR, NULL)"

(define-cproc make-u16vector (length::<fixnum> :optional (fill 0))
  (let* ([filler :: (unsigned short)])
    (u16unboxer filler fill)
    (return (Scm_MakeU16Vector length filler))))

(define-cproc u16vector (:optarray (elts nelts 10) :rest args)
  :fast-flonum
  (cond [(SCM_NULLP args)
         (return (Scm_ObjArrayToU16Vector elts nelts SCM_CLAMP_ERROR))]
        [else
         (let* ([i::int (- nelts 1)] [p args])
           (for [() (>= i 0) (post-- i)]
                (SCM_FLONUM_ENSURE_MEM (aref elts i))
                (set! p (Scm_Cons (aref elts i) p)))
           (return (Scm_ListToUVector SCM_CLASS_U16VECTOR p SCM_CLAMP_ERROR)))]))

(define-cproc u16vector-length (v::<u16vector>) ::<int> SCM_U16VECTOR_SIZE)

;; This is only called via apply.  The standard call will be inlined
;; as UVEC-REF instruction.
(define-cproc u16vector-ref
  (v::<u16vector> i::<fixnum> :optional fallback) :fast-flonum
  (setter u16vector-set!)
  (return (Scm_VMUVectorRef v (Scm_UVectorType (Scm_ClassOf (SCM_OBJ v)))
                            i fallback)))

(define-cproc u16vector-set!
  (v::<u16vector> i::<fixnum> val :optional clamp) :fast-flonum
  (return (Scm_U16VectorSet v i val (clamp_arg clamp))))

(define-cproc u16vector-copy
  (v::<u16vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_U16VectorCopy)

;; NB: this function accomodates to two APIs.
;; The 'modern' API, compatible to srfi-13 and srfi-43, takes arguments:
;;    (dst dstart src :optional sstart send)
;; The old API only takes:
;;    (dst src)
;; We dispatch by the second argument.  The old API is deprecated, but
;; kept for the existing code.
(define-cproc u16vector-copy!
  (dst::<u16vector> dstart :optional src (sstart::<int> 0) (send::<int> -1))
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (cond
   [(SCM_INTEGERP dstart) ; new API
    (unless (SCM_U16VECTORP src)
      (Scm_Error "u16vector-copy!: argument is not a matching uvector: %S"src))
    (return (Scm_U16VectorCopyX dst (Scm_GetInteger dstart)
                  (SCM_U16VECTOR src) sstart send))]
   [(SCM_U16VECTORP dstart) ; old API
    (return (Scm_U16VectorCopyX dst 0 (SCM_U16VECTOR dstart) 0 -1))]
   [else (SCM_TYPE_ERROR dstart "integer") (return SCM_UNDEFINED)]))

;; multi-copy! - copy src repeatedly to different parts of dst.
(define-cproc u16vector-multi-copy!
  (dst::<u16vector> dstart::<fixnum> dstride::<fixnum>
   src::<u16vector> :optional (sstart::<fixnum> 0) (ssize::<fixnum> 0)
                               (sstride::<fixnum> -1) (count::<fixnum> -1))
  ::<void>
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (when (< sstride 0)
    (if (<= ssize 0) (set! sstride 0) (set! sstride ssize)))
  (when (<= ssize 0)
    (set! ssize (- (SCM_UVECTOR_SIZE src) sstart)))
  (let* ([di::long dstart] [si::long sstart]
         [dlen::long (SCM_UVECTOR_SIZE dst)]
         [slen::long (SCM_UVECTOR_SIZE src)]
         [c::long 0] [k::long]
         [cnt::long (?: (< count 0) LONG_MAX count)])
    (if (== ssize 1)
      ;; shortcut for optimization
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (set! (SCM_U16VECTOR_ELEMENT dst di)
                 (SCM_U16VECTOR_ELEMENT src si))
           (+= di dstride)
           (+= si sstride))
      ;; general case
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (let* ([nitems::long ssize])
             (when (< dlen (+ di nitems)) (set! nitems (- dlen di)))
             (when (< slen (+ si nitems)) (set! nitems (- slen si)))
             (memmove (+ (SCM_U16VECTOR_ELEMENTS dst) di)
                      (+ (SCM_U16VECTOR_ELEMENTS src) si)
                      (* nitems (sizeof (.type unsigned short)))))
           (+= di dstride)
           (+= si sstride)))))

(define-cproc u16vector->list
  (v::<u16vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_U16VectorToList)

(define-cproc list->u16vector (l::<list> :optional clamp)
  (return (Scm_ListToUVector SCM_CLASS_U16VECTOR l (clamp-arg clamp))))

(define-cproc u16vector-fill!
  (v::<u16vector> val :optional (start::<fixnum> 0) (end::<fixnum> -1))
  (let* ([filler :: (unsigned short)])
    (u16unboxer filler val)
    (return (Scm_U16VectorFill v filler start end))))

(define-cproc u16vector->vector
  (v::<u16vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_U16VectorToVector)

(define-cproc vector->u16vector
  (v::<vector> :optional (start::<fixnum> 0) (end::<fixnum> -1) clamp)
  (return (Scm_VectorToU16Vector v start end (clamp-arg clamp))))


;;--------------------------------------------------------------------
;; s32vector basic API
;;
(define-cproc s32vector? (obj) ::<boolean> SCM_S32VECTORP)

;; we need this in string, for the UNBOX macro is substituted to the
;; literal C code instead of CISE code.
"#define s32unboxer(filler, fill) filler = (ScmInt32)Scm_GetInteger32Clamp(fill, SCM_CLAMP_ERROR, NULL)"

(define-cproc make-s32vector (length::<fixnum> :optional (fill 0))
  (let* ([filler :: (ScmInt32)])
    (s32unboxer filler fill)
    (return (Scm_MakeS32Vector length filler))))

(define-cproc s32vector (:optarray (elts nelts 10) :rest args)
  :fast-flonum
  (cond [(SCM_NULLP args)
         (return (Scm_ObjArrayToS32Vector elts nelts SCM_CLAMP_ERROR))]
        [else
         (let* ([i::int (- nelts 1)] [p args])
           (for [() (>= i 0) (post-- i)]
                (SCM_FLONUM_ENSURE_MEM (aref elts i))
                (set! p (Scm_Cons (aref elts i) p)))
           (return (Scm_ListToUVector SCM_CLASS_S32VECTOR p SCM_CLAMP_ERROR)))]))

(define-cproc s32vector-length (v::<s32vector>) ::<int> SCM_S32VECTOR_SIZE)

;; This is only called via apply.  The standard call will be inlined
;; as UVEC-REF instruction.
(define-cproc s32vector-ref
  (v::<s32vector> i::<fixnum> :optional fallback) :fast-flonum
  (setter s32vector-set!)
  (return (Scm_VMUVectorRef v (Scm_UVectorType (Scm_ClassOf (SCM_OBJ v)))
                            i fallback)))

(define-cproc s32vector-set!
  (v::<s32vector> i::<fixnum> val :optional clamp) :fast-flonum
  (return (Scm_S32VectorSet v i val (clamp_arg clamp))))

(define-cproc s32vector-copy
  (v::<s32vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_S32VectorCopy)

;; NB: this function accomodates to two APIs.
;; The 'modern' API, compatible to srfi-13 and srfi-43, takes arguments:
;;    (dst dstart src :optional sstart send)
;; The old API only takes:
;;    (dst src)
;; We dispatch by the second argument.  The old API is deprecated, but
;; kept for the existing code.
(define-cproc s32vector-copy!
  (dst::<s32vector> dstart :optional src (sstart::<int> 0) (send::<int> -1))
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (cond
   [(SCM_INTEGERP dstart) ; new API
    (unless (SCM_S32VECTORP src)
      (Scm_Error "s32vector-copy!: argument is not a matching uvector: %S"src))
    (return (Scm_S32VectorCopyX dst (Scm_GetInteger dstart)
                  (SCM_S32VECTOR src) sstart send))]
   [(SCM_S32VECTORP dstart) ; old API
    (return (Scm_S32VectorCopyX dst 0 (SCM_S32VECTOR dstart) 0 -1))]
   [else (SCM_TYPE_ERROR dstart "integer") (return SCM_UNDEFINED)]))

;; multi-copy! - copy src repeatedly to different parts of dst.
(define-cproc s32vector-multi-copy!
  (dst::<s32vector> dstart::<fixnum> dstride::<fixnum>
   src::<s32vector> :optional (sstart::<fixnum> 0) (ssize::<fixnum> 0)
                               (sstride::<fixnum> -1) (count::<fixnum> -1))
  ::<void>
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (when (< sstride 0)
    (if (<= ssize 0) (set! sstride 0) (set! sstride ssize)))
  (when (<= ssize 0)
    (set! ssize (- (SCM_UVECTOR_SIZE src) sstart)))
  (let* ([di::long dstart] [si::long sstart]
         [dlen::long (SCM_UVECTOR_SIZE dst)]
         [slen::long (SCM_UVECTOR_SIZE src)]
         [c::long 0] [k::long]
         [cnt::long (?: (< count 0) LONG_MAX count)])
    (if (== ssize 1)
      ;; shortcut for optimization
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (set! (SCM_S32VECTOR_ELEMENT dst di)
                 (SCM_S32VECTOR_ELEMENT src si))
           (+= di dstride)
           (+= si sstride))
      ;; general case
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (let* ([nitems::long ssize])
             (when (< dlen (+ di nitems)) (set! nitems (- dlen di)))
             (when (< slen (+ si nitems)) (set! nitems (- slen si)))
             (memmove (+ (SCM_S32VECTOR_ELEMENTS dst) di)
                      (+ (SCM_S32VECTOR_ELEMENTS src) si)
                      (* nitems (sizeof (.type ScmInt32)))))
           (+= di dstride)
           (+= si sstride)))))

(define-cproc s32vector->list
  (v::<s32vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_S32VectorToList)

(define-cproc list->s32vector (l::<list> :optional clamp)
  (return (Scm_ListToUVector SCM_CLASS_S32VECTOR l (clamp-arg clamp))))

(define-cproc s32vector-fill!
  (v::<s32vector> val :optional (start::<fixnum> 0) (end::<fixnum> -1))
  (let* ([filler :: (ScmInt32)])
    (s32unboxer filler val)
    (return (Scm_S32VectorFill v filler start end))))

(define-cproc s32vector->vector
  (v::<s32vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_S32VectorToVector)

(define-cproc vector->s32vector
  (v::<vector> :optional (start::<fixnum> 0) (end::<fixnum> -1) clamp)
  (return (Scm_VectorToS32Vector v start end (clamp-arg clamp))))


;;--------------------------------------------------------------------
;; u32vector basic API
;;
(define-cproc u32vector? (obj) ::<boolean> SCM_U32VECTORP)

;; we need this in string, for the UNBOX macro is substituted to the
;; literal C code instead of CISE code.
"#define u32unboxer(filler, fill) filler = (ScmUInt32)Scm_GetIntegerU32Clamp(fill, SCM_CLAMP_ERROR, NULL)"

(define-cproc make-u32vector (length::<fixnum> :optional (fill 0))
  (let* ([filler :: (ScmUInt32)])
    (u32unboxer filler fill)
    (return (Scm_MakeU32Vector length filler))))

(define-cproc u32vector (:optarray (elts nelts 10) :rest args)
  :fast-flonum
  (cond [(SCM_NULLP args)
         (return (Scm_ObjArrayToU32Vector elts nelts SCM_CLAMP_ERROR))]
        [else
         (let* ([i::int (- nelts 1)] [p args])
           (for [() (>= i 0) (post-- i)]
                (SCM_FLONUM_ENSURE_MEM (aref elts i))
                (set! p (Scm_Cons (aref elts i) p)))
           (return (Scm_ListToUVector SCM_CLASS_U32VECTOR p SCM_CLAMP_ERROR)))]))

(define-cproc u32vector-length (v::<u32vector>) ::<int> SCM_U32VECTOR_SIZE)

;; This is only called via apply.  The standard call will be inlined
;; as UVEC-REF instruction.
(define-cproc u32vector-ref
  (v::<u32vector> i::<fixnum> :optional fallback) :fast-flonum
  (setter u32vector-set!)
  (return (Scm_VMUVectorRef v (Scm_UVectorType (Scm_ClassOf (SCM_OBJ v)))
                            i fallback)))

(define-cproc u32vector-set!
  (v::<u32vector> i::<fixnum> val :optional clamp) :fast-flonum
  (return (Scm_U32VectorSet v i val (clamp_arg clamp))))

(define-cproc u32vector-copy
  (v::<u32vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_U32VectorCopy)

;; NB: this function accomodates to two APIs.
;; The 'modern' API, compatible to srfi-13 and srfi-43, takes arguments:
;;    (dst dstart src :optional sstart send)
;; The old API only takes:
;;    (dst src)
;; We dispatch by the second argument.  The old API is deprecated, but
;; kept for the existing code.
(define-cproc u32vector-copy!
  (dst::<u32vector> dstart :optional src (sstart::<int> 0) (send::<int> -1))
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (cond
   [(SCM_INTEGERP dstart) ; new API
    (unless (SCM_U32VECTORP src)
      (Scm_Error "u32vector-copy!: argument is not a matching uvector: %S"src))
    (return (Scm_U32VectorCopyX dst (Scm_GetInteger dstart)
                  (SCM_U32VECTOR src) sstart send))]
   [(SCM_U32VECTORP dstart) ; old API
    (return (Scm_U32VectorCopyX dst 0 (SCM_U32VECTOR dstart) 0 -1))]
   [else (SCM_TYPE_ERROR dstart "integer") (return SCM_UNDEFINED)]))

;; multi-copy! - copy src repeatedly to different parts of dst.
(define-cproc u32vector-multi-copy!
  (dst::<u32vector> dstart::<fixnum> dstride::<fixnum>
   src::<u32vector> :optional (sstart::<fixnum> 0) (ssize::<fixnum> 0)
                               (sstride::<fixnum> -1) (count::<fixnum> -1))
  ::<void>
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (when (< sstride 0)
    (if (<= ssize 0) (set! sstride 0) (set! sstride ssize)))
  (when (<= ssize 0)
    (set! ssize (- (SCM_UVECTOR_SIZE src) sstart)))
  (let* ([di::long dstart] [si::long sstart]
         [dlen::long (SCM_UVECTOR_SIZE dst)]
         [slen::long (SCM_UVECTOR_SIZE src)]
         [c::long 0] [k::long]
         [cnt::long (?: (< count 0) LONG_MAX count)])
    (if (== ssize 1)
      ;; shortcut for optimization
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (set! (SCM_U32VECTOR_ELEMENT dst di)
                 (SCM_U32VECTOR_ELEMENT src si))
           (+= di dstride)
           (+= si sstride))
      ;; general case
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (let* ([nitems::long ssize])
             (when (< dlen (+ di nitems)) (set! nitems (- dlen di)))
             (when (< slen (+ si nitems)) (set! nitems (- slen si)))
             (memmove (+ (SCM_U32VECTOR_ELEMENTS dst) di)
                      (+ (SCM_U32VECTOR_ELEMENTS src) si)
                      (* nitems (sizeof (.type ScmUInt32)))))
           (+= di dstride)
           (+= si sstride)))))

(define-cproc u32vector->list
  (v::<u32vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_U32VectorToList)

(define-cproc list->u32vector (l::<list> :optional clamp)
  (return (Scm_ListToUVector SCM_CLASS_U32VECTOR l (clamp-arg clamp))))

(define-cproc u32vector-fill!
  (v::<u32vector> val :optional (start::<fixnum> 0) (end::<fixnum> -1))
  (let* ([filler :: (ScmUInt32)])
    (u32unboxer filler val)
    (return (Scm_U32VectorFill v filler start end))))

(define-cproc u32vector->vector
  (v::<u32vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_U32VectorToVector)

(define-cproc vector->u32vector
  (v::<vector> :optional (start::<fixnum> 0) (end::<fixnum> -1) clamp)
  (return (Scm_VectorToU32Vector v start end (clamp-arg clamp))))


;;--------------------------------------------------------------------
;; s64vector basic API
;;
(define-cproc s64vector? (obj) ::<boolean> SCM_S64VECTORP)

;; we need this in string, for the UNBOX macro is substituted to the
;; literal C code instead of CISE code.
"#define s64unboxer(filler, fill) filler = Scm_GetInteger64Clamp(fill, SCM_CLAMP_ERROR, NULL)"

(define-cproc make-s64vector (length::<fixnum> :optional (fill 0))
  (let* ([filler :: (ScmInt64)])
    (s64unboxer filler fill)
    (return (Scm_MakeS64Vector length filler))))

(define-cproc s64vector (:optarray (elts nelts 10) :rest args)
  :fast-flonum
  (cond [(SCM_NULLP args)
         (return (Scm_ObjArrayToS64Vector elts nelts SCM_CLAMP_ERROR))]
        [else
         (let* ([i::int (- nelts 1)] [p args])
           (for [() (>= i 0) (post-- i)]
                (SCM_FLONUM_ENSURE_MEM (aref elts i))
                (set! p (Scm_Cons (aref elts i) p)))
           (return (Scm_ListToUVector SCM_CLASS_S64VECTOR p SCM_CLAMP_ERROR)))]))

(define-cproc s64vector-length (v::<s64vector>) ::<int> SCM_S64VECTOR_SIZE)

;; This is only called via apply.  The standard call will be inlined
;; as UVEC-REF instruction.
(define-cproc s64vector-ref
  (v::<s64vector> i::<fixnum> :optional fallback) :fast-flonum
  (setter s64vector-set!)
  (return (Scm_VMUVectorRef v (Scm_UVectorType (Scm_ClassOf (SCM_OBJ v)))
                            i fallback)))

(define-cproc s64vector-set!
  (v::<s64vector> i::<fixnum> val :optional clamp) :fast-flonum
  (return (Scm_S64VectorSet v i val (clamp_arg clamp))))

(define-cproc s64vector-copy
  (v::<s64vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_S64VectorCopy)

;; NB: this function accomodates to two APIs.
;; The 'modern' API, compatible to srfi-13 and srfi-43, takes arguments:
;;    (dst dstart src :optional sstart send)
;; The old API only takes:
;;    (dst src)
;; We dispatch by the second argument.  The old API is deprecated, but
;; kept for the existing code.
(define-cproc s64vector-copy!
  (dst::<s64vector> dstart :optional src (sstart::<int> 0) (send::<int> -1))
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (cond
   [(SCM_INTEGERP dstart) ; new API
    (unless (SCM_S64VECTORP src)
      (Scm_Error "s64vector-copy!: argument is not a matching uvector: %S"src))
    (return (Scm_S64VectorCopyX dst (Scm_GetInteger dstart)
                  (SCM_S64VECTOR src) sstart send))]
   [(SCM_S64VECTORP dstart) ; old API
    (return (Scm_S64VectorCopyX dst 0 (SCM_S64VECTOR dstart) 0 -1))]
   [else (SCM_TYPE_ERROR dstart "integer") (return SCM_UNDEFINED)]))

;; multi-copy! - copy src repeatedly to different parts of dst.
(define-cproc s64vector-multi-copy!
  (dst::<s64vector> dstart::<fixnum> dstride::<fixnum>
   src::<s64vector> :optional (sstart::<fixnum> 0) (ssize::<fixnum> 0)
                               (sstride::<fixnum> -1) (count::<fixnum> -1))
  ::<void>
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (when (< sstride 0)
    (if (<= ssize 0) (set! sstride 0) (set! sstride ssize)))
  (when (<= ssize 0)
    (set! ssize (- (SCM_UVECTOR_SIZE src) sstart)))
  (let* ([di::long dstart] [si::long sstart]
         [dlen::long (SCM_UVECTOR_SIZE dst)]
         [slen::long (SCM_UVECTOR_SIZE src)]
         [c::long 0] [k::long]
         [cnt::long (?: (< count 0) LONG_MAX count)])
    (if (== ssize 1)
      ;; shortcut for optimization
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (set! (SCM_S64VECTOR_ELEMENT dst di)
                 (SCM_S64VECTOR_ELEMENT src si))
           (+= di dstride)
           (+= si sstride))
      ;; general case
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (let* ([nitems::long ssize])
             (when (< dlen (+ di nitems)) (set! nitems (- dlen di)))
             (when (< slen (+ si nitems)) (set! nitems (- slen si)))
             (memmove (+ (SCM_S64VECTOR_ELEMENTS dst) di)
                      (+ (SCM_S64VECTOR_ELEMENTS src) si)
                      (* nitems (sizeof (.type ScmInt64)))))
           (+= di dstride)
           (+= si sstride)))))

(define-cproc s64vector->list
  (v::<s64vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_S64VectorToList)

(define-cproc list->s64vector (l::<list> :optional clamp)
  (return (Scm_ListToUVector SCM_CLASS_S64VECTOR l (clamp-arg clamp))))

(define-cproc s64vector-fill!
  (v::<s64vector> val :optional (start::<fixnum> 0) (end::<fixnum> -1))
  (let* ([filler :: (ScmInt64)])
    (s64unboxer filler val)
    (return (Scm_S64VectorFill v filler start end))))

(define-cproc s64vector->vector
  (v::<s64vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_S64VectorToVector)

(define-cproc vector->s64vector
  (v::<vector> :optional (start::<fixnum> 0) (end::<fixnum> -1) clamp)
  (return (Scm_VectorToS64Vector v start end (clamp-arg clamp))))


;;--------------------------------------------------------------------
;; u64vector basic API
;;
(define-cproc u64vector? (obj) ::<boolean> SCM_U64VECTORP)

;; we need this in string, for the UNBOX macro is substituted to the
;; literal C code instead of CISE code.
"#define u64unboxer(filler, fill) filler = Scm_GetIntegerU64Clamp(fill, SCM_CLAMP_ERROR, NULL)"

(define-cproc make-u64vector (length::<fixnum> :optional (fill 0))
  (let* ([filler :: (ScmUInt64)])
    (u64unboxer filler fill)
    (return (Scm_MakeU64Vector length filler))))

(define-cproc u64vector (:optarray (elts nelts 10) :rest args)
  :fast-flonum
  (cond [(SCM_NULLP args)
         (return (Scm_ObjArrayToU64Vector elts nelts SCM_CLAMP_ERROR))]
        [else
         (let* ([i::int (- nelts 1)] [p args])
           (for [() (>= i 0) (post-- i)]
                (SCM_FLONUM_ENSURE_MEM (aref elts i))
                (set! p (Scm_Cons (aref elts i) p)))
           (return (Scm_ListToUVector SCM_CLASS_U64VECTOR p SCM_CLAMP_ERROR)))]))

(define-cproc u64vector-length (v::<u64vector>) ::<int> SCM_U64VECTOR_SIZE)

;; This is only called via apply.  The standard call will be inlined
;; as UVEC-REF instruction.
(define-cproc u64vector-ref
  (v::<u64vector> i::<fixnum> :optional fallback) :fast-flonum
  (setter u64vector-set!)
  (return (Scm_VMUVectorRef v (Scm_UVectorType (Scm_ClassOf (SCM_OBJ v)))
                            i fallback)))

(define-cproc u64vector-set!
  (v::<u64vector> i::<fixnum> val :optional clamp) :fast-flonum
  (return (Scm_U64VectorSet v i val (clamp_arg clamp))))

(define-cproc u64vector-copy
  (v::<u64vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_U64VectorCopy)

;; NB: this function accomodates to two APIs.
;; The 'modern' API, compatible to srfi-13 and srfi-43, takes arguments:
;;    (dst dstart src :optional sstart send)
;; The old API only takes:
;;    (dst src)
;; We dispatch by the second argument.  The old API is deprecated, but
;; kept for the existing code.
(define-cproc u64vector-copy!
  (dst::<u64vector> dstart :optional src (sstart::<int> 0) (send::<int> -1))
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (cond
   [(SCM_INTEGERP dstart) ; new API
    (unless (SCM_U64VECTORP src)
      (Scm_Error "u64vector-copy!: argument is not a matching uvector: %S"src))
    (return (Scm_U64VectorCopyX dst (Scm_GetInteger dstart)
                  (SCM_U64VECTOR src) sstart send))]
   [(SCM_U64VECTORP dstart) ; old API
    (return (Scm_U64VectorCopyX dst 0 (SCM_U64VECTOR dstart) 0 -1))]
   [else (SCM_TYPE_ERROR dstart "integer") (return SCM_UNDEFINED)]))

;; multi-copy! - copy src repeatedly to different parts of dst.
(define-cproc u64vector-multi-copy!
  (dst::<u64vector> dstart::<fixnum> dstride::<fixnum>
   src::<u64vector> :optional (sstart::<fixnum> 0) (ssize::<fixnum> 0)
                               (sstride::<fixnum> -1) (count::<fixnum> -1))
  ::<void>
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (when (< sstride 0)
    (if (<= ssize 0) (set! sstride 0) (set! sstride ssize)))
  (when (<= ssize 0)
    (set! ssize (- (SCM_UVECTOR_SIZE src) sstart)))
  (let* ([di::long dstart] [si::long sstart]
         [dlen::long (SCM_UVECTOR_SIZE dst)]
         [slen::long (SCM_UVECTOR_SIZE src)]
         [c::long 0] [k::long]
         [cnt::long (?: (< count 0) LONG_MAX count)])
    (if (== ssize 1)
      ;; shortcut for optimization
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (set! (SCM_U64VECTOR_ELEMENT dst di)
                 (SCM_U64VECTOR_ELEMENT src si))
           (+= di dstride)
           (+= si sstride))
      ;; general case
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (let* ([nitems::long ssize])
             (when (< dlen (+ di nitems)) (set! nitems (- dlen di)))
             (when (< slen (+ si nitems)) (set! nitems (- slen si)))
             (memmove (+ (SCM_U64VECTOR_ELEMENTS dst) di)
                      (+ (SCM_U64VECTOR_ELEMENTS src) si)
                      (* nitems (sizeof (.type ScmUInt64)))))
           (+= di dstride)
           (+= si sstride)))))

(define-cproc u64vector->list
  (v::<u64vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_U64VectorToList)

(define-cproc list->u64vector (l::<list> :optional clamp)
  (return (Scm_ListToUVector SCM_CLASS_U64VECTOR l (clamp-arg clamp))))

(define-cproc u64vector-fill!
  (v::<u64vector> val :optional (start::<fixnum> 0) (end::<fixnum> -1))
  (let* ([filler :: (ScmUInt64)])
    (u64unboxer filler val)
    (return (Scm_U64VectorFill v filler start end))))

(define-cproc u64vector->vector
  (v::<u64vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_U64VectorToVector)

(define-cproc vector->u64vector
  (v::<vector> :optional (start::<fixnum> 0) (end::<fixnum> -1) clamp)
  (return (Scm_VectorToU64Vector v start end (clamp-arg clamp))))


;;--------------------------------------------------------------------
;; f16vector basic API
;;
(define-cproc f16vector? (obj) ::<boolean> SCM_F16VECTORP)

;; we need this in string, for the UNBOX macro is substituted to the
;; literal C code instead of CISE code.
"#define f16unboxer(filler, fill) filler = Scm_DoubleToHalf(Scm_GetDouble(fill))"

(define-cproc make-f16vector (length::<fixnum> :optional (fill 0))
  (let* ([filler :: (ScmHalfFloat)])
    (f16unboxer filler fill)
    (return (Scm_MakeF16Vector length filler))))

(define-cproc f16vector (:optarray (elts nelts 10) :rest args)
  :fast-flonum
  (cond [(SCM_NULLP args)
         (return (Scm_ObjArrayToF16Vector elts nelts SCM_CLAMP_ERROR))]
        [else
         (let* ([i::int (- nelts 1)] [p args])
           (for [() (>= i 0) (post-- i)]
                (SCM_FLONUM_ENSURE_MEM (aref elts i))
                (set! p (Scm_Cons (aref elts i) p)))
           (return (Scm_ListToUVector SCM_CLASS_F16VECTOR p SCM_CLAMP_ERROR)))]))

(define-cproc f16vector-length (v::<f16vector>) ::<int> SCM_F16VECTOR_SIZE)

;; This is only called via apply.  The standard call will be inlined
;; as UVEC-REF instruction.
(define-cproc f16vector-ref
  (v::<f16vector> i::<fixnum> :optional fallback) :fast-flonum
  (setter f16vector-set!)
  (return (Scm_VMUVectorRef v (Scm_UVectorType (Scm_ClassOf (SCM_OBJ v)))
                            i fallback)))

(define-cproc f16vector-set!
  (v::<f16vector> i::<fixnum> val :optional clamp) :fast-flonum
  (return (Scm_F16VectorSet v i val (clamp_arg clamp))))

(define-cproc f16vector-copy
  (v::<f16vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_F16VectorCopy)

;; NB: this function accomodates to two APIs.
;; The 'modern' API, compatible to srfi-13 and srfi-43, takes arguments:
;;    (dst dstart src :optional sstart send)
;; The old API only takes:
;;    (dst src)
;; We dispatch by the second argument.  The old API is deprecated, but
;; kept for the existing code.
(define-cproc f16vector-copy!
  (dst::<f16vector> dstart :optional src (sstart::<int> 0) (send::<int> -1))
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (cond
   [(SCM_INTEGERP dstart) ; new API
    (unless (SCM_F16VECTORP src)
      (Scm_Error "f16vector-copy!: argument is not a matching uvector: %S"src))
    (return (Scm_F16VectorCopyX dst (Scm_GetInteger dstart)
                  (SCM_F16VECTOR src) sstart send))]
   [(SCM_F16VECTORP dstart) ; old API
    (return (Scm_F16VectorCopyX dst 0 (SCM_F16VECTOR dstart) 0 -1))]
   [else (SCM_TYPE_ERROR dstart "integer") (return SCM_UNDEFINED)]))

;; multi-copy! - copy src repeatedly to different parts of dst.
(define-cproc f16vector-multi-copy!
  (dst::<f16vector> dstart::<fixnum> dstride::<fixnum>
   src::<f16vector> :optional (sstart::<fixnum> 0) (ssize::<fixnum> 0)
                               (sstride::<fixnum> -1) (count::<fixnum> -1))
  ::<void>
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (when (< sstride 0)
    (if (<= ssize 0) (set! sstride 0) (set! sstride ssize)))
  (when (<= ssize 0)
    (set! ssize (- (SCM_UVECTOR_SIZE src) sstart)))
  (let* ([di::long dstart] [si::long sstart]
         [dlen::long (SCM_UVECTOR_SIZE dst)]
         [slen::long (SCM_UVECTOR_SIZE src)]
         [c::long 0] [k::long]
         [cnt::long (?: (< count 0) LONG_MAX count)])
    (if (== ssize 1)
      ;; shortcut for optimization
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (set! (SCM_F16VECTOR_ELEMENT dst di)
                 (SCM_F16VECTOR_ELEMENT src si))
           (+= di dstride)
           (+= si sstride))
      ;; general case
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (let* ([nitems::long ssize])
             (when (< dlen (+ di nitems)) (set! nitems (- dlen di)))
             (when (< slen (+ si nitems)) (set! nitems (- slen si)))
             (memmove (+ (SCM_F16VECTOR_ELEMENTS dst) di)
                      (+ (SCM_F16VECTOR_ELEMENTS src) si)
                      (* nitems (sizeof (.type ScmHalfFloat)))))
           (+= di dstride)
           (+= si sstride)))))

(define-cproc f16vector->list
  (v::<f16vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_F16VectorToList)

(define-cproc list->f16vector (l::<list> :optional clamp)
  (return (Scm_ListToUVector SCM_CLASS_F16VECTOR l (clamp-arg clamp))))

(define-cproc f16vector-fill!
  (v::<f16vector> val :optional (start::<fixnum> 0) (end::<fixnum> -1))
  (let* ([filler :: (ScmHalfFloat)])
    (f16unboxer filler val)
    (return (Scm_F16VectorFill v filler start end))))

(define-cproc f16vector->vector
  (v::<f16vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_F16VectorToVector)

(define-cproc vector->f16vector
  (v::<vector> :optional (start::<fixnum> 0) (end::<fixnum> -1) clamp)
  (return (Scm_VectorToF16Vector v start end (clamp-arg clamp))))


;;--------------------------------------------------------------------
;; f32vector basic API
;;
(define-cproc f32vector? (obj) ::<boolean> SCM_F32VECTORP)

;; we need this in string, for the UNBOX macro is substituted to the
;; literal C code instead of CISE code.
"#define f32unboxer(filler, fill) filler = (float)Scm_GetDouble(fill)"

(define-cproc make-f32vector (length::<fixnum> :optional (fill 0))
  (let* ([filler :: (float)])
    (f32unboxer filler fill)
    (return (Scm_MakeF32Vector length filler))))

(define-cproc f32vector (:optarray (elts nelts 10) :rest args)
  :fast-flonum
  (cond [(SCM_NULLP args)
         (return (Scm_ObjArrayToF32Vector elts nelts SCM_CLAMP_ERROR))]
        [else
         (let* ([i::int (- nelts 1)] [p args])
           (for [() (>= i 0) (post-- i)]
                (SCM_FLONUM_ENSURE_MEM (aref elts i))
                (set! p (Scm_Cons (aref elts i) p)))
           (return (Scm_ListToUVector SCM_CLASS_F32VECTOR p SCM_CLAMP_ERROR)))]))

(define-cproc f32vector-length (v::<f32vector>) ::<int> SCM_F32VECTOR_SIZE)

;; This is only called via apply.  The standard call will be inlined
;; as UVEC-REF instruction.
(define-cproc f32vector-ref
  (v::<f32vector> i::<fixnum> :optional fallback) :fast-flonum
  (setter f32vector-set!)
  (return (Scm_VMUVectorRef v (Scm_UVectorType (Scm_ClassOf (SCM_OBJ v)))
                            i fallback)))

(define-cproc f32vector-set!
  (v::<f32vector> i::<fixnum> val :optional clamp) :fast-flonum
  (return (Scm_F32VectorSet v i val (clamp_arg clamp))))

(define-cproc f32vector-copy
  (v::<f32vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_F32VectorCopy)

;; NB: this function accomodates to two APIs.
;; The 'modern' API, compatible to srfi-13 and srfi-43, takes arguments:
;;    (dst dstart src :optional sstart send)
;; The old API only takes:
;;    (dst src)
;; We dispatch by the second argument.  The old API is deprecated, but
;; kept for the existing code.
(define-cproc f32vector-copy!
  (dst::<f32vector> dstart :optional src (sstart::<int> 0) (send::<int> -1))
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (cond
   [(SCM_INTEGERP dstart) ; new API
    (unless (SCM_F32VECTORP src)
      (Scm_Error "f32vector-copy!: argument is not a matching uvector: %S"src))
    (return (Scm_F32VectorCopyX dst (Scm_GetInteger dstart)
                  (SCM_F32VECTOR src) sstart send))]
   [(SCM_F32VECTORP dstart) ; old API
    (return (Scm_F32VectorCopyX dst 0 (SCM_F32VECTOR dstart) 0 -1))]
   [else (SCM_TYPE_ERROR dstart "integer") (return SCM_UNDEFINED)]))

;; multi-copy! - copy src repeatedly to different parts of dst.
(define-cproc f32vector-multi-copy!
  (dst::<f32vector> dstart::<fixnum> dstride::<fixnum>
   src::<f32vector> :optional (sstart::<fixnum> 0) (ssize::<fixnum> 0)
                               (sstride::<fixnum> -1) (count::<fixnum> -1))
  ::<void>
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (when (< sstride 0)
    (if (<= ssize 0) (set! sstride 0) (set! sstride ssize)))
  (when (<= ssize 0)
    (set! ssize (- (SCM_UVECTOR_SIZE src) sstart)))
  (let* ([di::long dstart] [si::long sstart]
         [dlen::long (SCM_UVECTOR_SIZE dst)]
         [slen::long (SCM_UVECTOR_SIZE src)]
         [c::long 0] [k::long]
         [cnt::long (?: (< count 0) LONG_MAX count)])
    (if (== ssize 1)
      ;; shortcut for optimization
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (set! (SCM_F32VECTOR_ELEMENT dst di)
                 (SCM_F32VECTOR_ELEMENT src si))
           (+= di dstride)
           (+= si sstride))
      ;; general case
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (let* ([nitems::long ssize])
             (when (< dlen (+ di nitems)) (set! nitems (- dlen di)))
             (when (< slen (+ si nitems)) (set! nitems (- slen si)))
             (memmove (+ (SCM_F32VECTOR_ELEMENTS dst) di)
                      (+ (SCM_F32VECTOR_ELEMENTS src) si)
                      (* nitems (sizeof (.type float)))))
           (+= di dstride)
           (+= si sstride)))))

(define-cproc f32vector->list
  (v::<f32vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_F32VectorToList)

(define-cproc list->f32vector (l::<list> :optional clamp)
  (return (Scm_ListToUVector SCM_CLASS_F32VECTOR l (clamp-arg clamp))))

(define-cproc f32vector-fill!
  (v::<f32vector> val :optional (start::<fixnum> 0) (end::<fixnum> -1))
  (let* ([filler :: (float)])
    (f32unboxer filler val)
    (return (Scm_F32VectorFill v filler start end))))

(define-cproc f32vector->vector
  (v::<f32vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_F32VectorToVector)

(define-cproc vector->f32vector
  (v::<vector> :optional (start::<fixnum> 0) (end::<fixnum> -1) clamp)
  (return (Scm_VectorToF32Vector v start end (clamp-arg clamp))))


;;--------------------------------------------------------------------
;; f64vector basic API
;;
(define-cproc f64vector? (obj) ::<boolean> SCM_F64VECTORP)

;; we need this in string, for the UNBOX macro is substituted to the
;; literal C code instead of CISE code.
"#define f64unboxer(filler, fill) filler = Scm_GetDouble(fill)"

(define-cproc make-f64vector (length::<fixnum> :optional (fill 0))
  (let* ([filler :: (double)])
    (f64unboxer filler fill)
    (return (Scm_MakeF64Vector length filler))))

(define-cproc f64vector (:optarray (elts nelts 10) :rest args)
  :fast-flonum
  (cond [(SCM_NULLP args)
         (return (Scm_ObjArrayToF64Vector elts nelts SCM_CLAMP_ERROR))]
        [else
         (let* ([i::int (- nelts 1)] [p args])
           (for [() (>= i 0) (post-- i)]
                (SCM_FLONUM_ENSURE_MEM (aref elts i))
                (set! p (Scm_Cons (aref elts i) p)))
           (return (Scm_ListToUVector SCM_CLASS_F64VECTOR p SCM_CLAMP_ERROR)))]))

(define-cproc f64vector-length (v::<f64vector>) ::<int> SCM_F64VECTOR_SIZE)

;; This is only called via apply.  The standard call will be inlined
;; as UVEC-REF instruction.
(define-cproc f64vector-ref
  (v::<f64vector> i::<fixnum> :optional fallback) :fast-flonum
  (setter f64vector-set!)
  (return (Scm_VMUVectorRef v (Scm_UVectorType (Scm_ClassOf (SCM_OBJ v)))
                            i fallback)))

(define-cproc f64vector-set!
  (v::<f64vector> i::<fixnum> val :optional clamp) :fast-flonum
  (return (Scm_F64VectorSet v i val (clamp_arg clamp))))

(define-cproc f64vector-copy
  (v::<f64vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_F64VectorCopy)

;; NB: this function accomodates to two APIs.
;; The 'modern' API, compatible to srfi-13 and srfi-43, takes arguments:
;;    (dst dstart src :optional sstart send)
;; The old API only takes:
;;    (dst src)
;; We dispatch by the second argument.  The old API is deprecated, but
;; kept for the existing code.
(define-cproc f64vector-copy!
  (dst::<f64vector> dstart :optional src (sstart::<int> 0) (send::<int> -1))
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (cond
   [(SCM_INTEGERP dstart) ; new API
    (unless (SCM_F64VECTORP src)
      (Scm_Error "f64vector-copy!: argument is not a matching uvector: %S"src))
    (return (Scm_F64VectorCopyX dst (Scm_GetInteger dstart)
                  (SCM_F64VECTOR src) sstart send))]
   [(SCM_F64VECTORP dstart) ; old API
    (return (Scm_F64VectorCopyX dst 0 (SCM_F64VECTOR dstart) 0 -1))]
   [else (SCM_TYPE_ERROR dstart "integer") (return SCM_UNDEFINED)]))

;; multi-copy! - copy src repeatedly to different parts of dst.
(define-cproc f64vector-multi-copy!
  (dst::<f64vector> dstart::<fixnum> dstride::<fixnum>
   src::<f64vector> :optional (sstart::<fixnum> 0) (ssize::<fixnum> 0)
                               (sstride::<fixnum> -1) (count::<fixnum> -1))
  ::<void>
  (SCM_UVECTOR_CHECK_MUTABLE dst)
  (when (< sstride 0)
    (if (<= ssize 0) (set! sstride 0) (set! sstride ssize)))
  (when (<= ssize 0)
    (set! ssize (- (SCM_UVECTOR_SIZE src) sstart)))
  (let* ([di::long dstart] [si::long sstart]
         [dlen::long (SCM_UVECTOR_SIZE dst)]
         [slen::long (SCM_UVECTOR_SIZE src)]
         [c::long 0] [k::long]
         [cnt::long (?: (< count 0) LONG_MAX count)])
    (if (== ssize 1)
      ;; shortcut for optimization
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (set! (SCM_F64VECTOR_ELEMENT dst di)
                 (SCM_F64VECTOR_ELEMENT src si))
           (+= di dstride)
           (+= si sstride))
      ;; general case
      (for [() (and (< c cnt) (< di dlen) (< si slen)) (inc! c)]
           (let* ([nitems::long ssize])
             (when (< dlen (+ di nitems)) (set! nitems (- dlen di)))
             (when (< slen (+ si nitems)) (set! nitems (- slen si)))
             (memmove (+ (SCM_F64VECTOR_ELEMENTS dst) di)
                      (+ (SCM_F64VECTOR_ELEMENTS src) si)
                      (* nitems (sizeof (.type double)))))
           (+= di dstride)
           (+= si sstride)))))

(define-cproc f64vector->list
  (v::<f64vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_F64VectorToList)

(define-cproc list->f64vector (l::<list> :optional clamp)
  (return (Scm_ListToUVector SCM_CLASS_F64VECTOR l (clamp-arg clamp))))

(define-cproc f64vector-fill!
  (v::<f64vector> val :optional (start::<fixnum> 0) (end::<fixnum> -1))
  (let* ([filler :: (double)])
    (f64unboxer filler val)
    (return (Scm_F64VectorFill v filler start end))))

(define-cproc f64vector->vector
  (v::<f64vector> :optional (start::<fixnum> 0) (end::<fixnum> -1))
  Scm_F64VectorToVector)

(define-cproc vector->f64vector
  (v::<vector> :optional (start::<fixnum> 0) (end::<fixnum> -1) clamp)
  (return (Scm_VectorToF64Vector v start end (clamp-arg clamp))))

(define-cproc s8vector-add (v0::<s8vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S8VectorAdd v0 v1 (clamp-arg clamp))))

(define-cproc s8vector-add! (v0::<s8vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S8VectorAddX v0 v1 (clamp-arg clamp))))
(define-cproc u8vector-add (v0::<u8vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U8VectorAdd v0 v1 (clamp-arg clamp))))

(define-cproc u8vector-add! (v0::<u8vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U8VectorAddX v0 v1 (clamp-arg clamp))))
(define-cproc s16vector-add (v0::<s16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S16VectorAdd v0 v1 (clamp-arg clamp))))

(define-cproc s16vector-add! (v0::<s16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S16VectorAddX v0 v1 (clamp-arg clamp))))
(define-cproc u16vector-add (v0::<u16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U16VectorAdd v0 v1 (clamp-arg clamp))))

(define-cproc u16vector-add! (v0::<u16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U16VectorAddX v0 v1 (clamp-arg clamp))))
(define-cproc s32vector-add (v0::<s32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S32VectorAdd v0 v1 (clamp-arg clamp))))

(define-cproc s32vector-add! (v0::<s32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S32VectorAddX v0 v1 (clamp-arg clamp))))
(define-cproc u32vector-add (v0::<u32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U32VectorAdd v0 v1 (clamp-arg clamp))))

(define-cproc u32vector-add! (v0::<u32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U32VectorAddX v0 v1 (clamp-arg clamp))))
(define-cproc s64vector-add (v0::<s64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S64VectorAdd v0 v1 (clamp-arg clamp))))

(define-cproc s64vector-add! (v0::<s64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S64VectorAddX v0 v1 (clamp-arg clamp))))
(define-cproc u64vector-add (v0::<u64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U64VectorAdd v0 v1 (clamp-arg clamp))))

(define-cproc u64vector-add! (v0::<u64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U64VectorAddX v0 v1 (clamp-arg clamp))))
(define-cproc f16vector-add (v0::<f16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F16VectorAdd v0 v1 (clamp-arg clamp))))

(define-cproc f16vector-add! (v0::<f16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F16VectorAddX v0 v1 (clamp-arg clamp))))
(define-cproc f32vector-add (v0::<f32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F32VectorAdd v0 v1 (clamp-arg clamp))))

(define-cproc f32vector-add! (v0::<f32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F32VectorAddX v0 v1 (clamp-arg clamp))))
(define-cproc f64vector-add (v0::<f64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F64VectorAdd v0 v1 (clamp-arg clamp))))

(define-cproc f64vector-add! (v0::<f64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F64VectorAddX v0 v1 (clamp-arg clamp))))
(define-cproc s8vector-sub (v0::<s8vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S8VectorSub v0 v1 (clamp-arg clamp))))

(define-cproc s8vector-sub! (v0::<s8vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S8VectorSubX v0 v1 (clamp-arg clamp))))
(define-cproc u8vector-sub (v0::<u8vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U8VectorSub v0 v1 (clamp-arg clamp))))

(define-cproc u8vector-sub! (v0::<u8vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U8VectorSubX v0 v1 (clamp-arg clamp))))
(define-cproc s16vector-sub (v0::<s16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S16VectorSub v0 v1 (clamp-arg clamp))))

(define-cproc s16vector-sub! (v0::<s16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S16VectorSubX v0 v1 (clamp-arg clamp))))
(define-cproc u16vector-sub (v0::<u16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U16VectorSub v0 v1 (clamp-arg clamp))))

(define-cproc u16vector-sub! (v0::<u16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U16VectorSubX v0 v1 (clamp-arg clamp))))
(define-cproc s32vector-sub (v0::<s32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S32VectorSub v0 v1 (clamp-arg clamp))))

(define-cproc s32vector-sub! (v0::<s32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S32VectorSubX v0 v1 (clamp-arg clamp))))
(define-cproc u32vector-sub (v0::<u32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U32VectorSub v0 v1 (clamp-arg clamp))))

(define-cproc u32vector-sub! (v0::<u32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U32VectorSubX v0 v1 (clamp-arg clamp))))
(define-cproc s64vector-sub (v0::<s64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S64VectorSub v0 v1 (clamp-arg clamp))))

(define-cproc s64vector-sub! (v0::<s64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S64VectorSubX v0 v1 (clamp-arg clamp))))
(define-cproc u64vector-sub (v0::<u64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U64VectorSub v0 v1 (clamp-arg clamp))))

(define-cproc u64vector-sub! (v0::<u64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U64VectorSubX v0 v1 (clamp-arg clamp))))
(define-cproc f16vector-sub (v0::<f16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F16VectorSub v0 v1 (clamp-arg clamp))))

(define-cproc f16vector-sub! (v0::<f16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F16VectorSubX v0 v1 (clamp-arg clamp))))
(define-cproc f32vector-sub (v0::<f32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F32VectorSub v0 v1 (clamp-arg clamp))))

(define-cproc f32vector-sub! (v0::<f32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F32VectorSubX v0 v1 (clamp-arg clamp))))
(define-cproc f64vector-sub (v0::<f64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F64VectorSub v0 v1 (clamp-arg clamp))))

(define-cproc f64vector-sub! (v0::<f64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F64VectorSubX v0 v1 (clamp-arg clamp))))
(define-cproc s8vector-mul (v0::<s8vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S8VectorMul v0 v1 (clamp-arg clamp))))

(define-cproc s8vector-mul! (v0::<s8vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S8VectorMulX v0 v1 (clamp-arg clamp))))
(define-cproc u8vector-mul (v0::<u8vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U8VectorMul v0 v1 (clamp-arg clamp))))

(define-cproc u8vector-mul! (v0::<u8vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U8VectorMulX v0 v1 (clamp-arg clamp))))
(define-cproc s16vector-mul (v0::<s16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S16VectorMul v0 v1 (clamp-arg clamp))))

(define-cproc s16vector-mul! (v0::<s16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S16VectorMulX v0 v1 (clamp-arg clamp))))
(define-cproc u16vector-mul (v0::<u16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U16VectorMul v0 v1 (clamp-arg clamp))))

(define-cproc u16vector-mul! (v0::<u16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U16VectorMulX v0 v1 (clamp-arg clamp))))
(define-cproc s32vector-mul (v0::<s32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S32VectorMul v0 v1 (clamp-arg clamp))))

(define-cproc s32vector-mul! (v0::<s32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S32VectorMulX v0 v1 (clamp-arg clamp))))
(define-cproc u32vector-mul (v0::<u32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U32VectorMul v0 v1 (clamp-arg clamp))))

(define-cproc u32vector-mul! (v0::<u32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U32VectorMulX v0 v1 (clamp-arg clamp))))
(define-cproc s64vector-mul (v0::<s64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S64VectorMul v0 v1 (clamp-arg clamp))))

(define-cproc s64vector-mul! (v0::<s64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_S64VectorMulX v0 v1 (clamp-arg clamp))))
(define-cproc u64vector-mul (v0::<u64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U64VectorMul v0 v1 (clamp-arg clamp))))

(define-cproc u64vector-mul! (v0::<u64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_U64VectorMulX v0 v1 (clamp-arg clamp))))
(define-cproc f16vector-mul (v0::<f16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F16VectorMul v0 v1 (clamp-arg clamp))))

(define-cproc f16vector-mul! (v0::<f16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F16VectorMulX v0 v1 (clamp-arg clamp))))
(define-cproc f32vector-mul (v0::<f32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F32VectorMul v0 v1 (clamp-arg clamp))))

(define-cproc f32vector-mul! (v0::<f32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F32VectorMulX v0 v1 (clamp-arg clamp))))
(define-cproc f64vector-mul (v0::<f64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F64VectorMul v0 v1 (clamp-arg clamp))))

(define-cproc f64vector-mul! (v0::<f64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F64VectorMulX v0 v1 (clamp-arg clamp))))
(define-cproc f16vector-div (v0::<f16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F16VectorDiv v0 v1 (clamp-arg clamp))))

(define-cproc f16vector-div! (v0::<f16vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F16VectorDivX v0 v1 (clamp-arg clamp))))
(define-cproc f32vector-div (v0::<f32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F32VectorDiv v0 v1 (clamp-arg clamp))))

(define-cproc f32vector-div! (v0::<f32vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F32VectorDivX v0 v1 (clamp-arg clamp))))
(define-cproc f64vector-div (v0::<f64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F64VectorDiv v0 v1 (clamp-arg clamp))))

(define-cproc f64vector-div! (v0::<f64vector> v1 :optional clamp)
  :fast-flonum
  (return (Scm_F64VectorDivX v0 v1 (clamp-arg clamp))))
(define-cproc s8vector-and (v0::<s8vector> v1)
  (return (Scm_S8VectorAnd v0 v1)))

(define-cproc s8vector-and! (v0::<s8vector> v1)
  (return (Scm_S8VectorAndX v0 v1)))
(define-cproc s8vector-ior (v0::<s8vector> v1)
  (return (Scm_S8VectorIor v0 v1)))

(define-cproc s8vector-ior! (v0::<s8vector> v1)
  (return (Scm_S8VectorIorX v0 v1)))
(define-cproc s8vector-xor (v0::<s8vector> v1)
  (return (Scm_S8VectorXor v0 v1)))

(define-cproc s8vector-xor! (v0::<s8vector> v1)
  (return (Scm_S8VectorXorX v0 v1)))
(define-cproc u8vector-and (v0::<u8vector> v1)
  (return (Scm_U8VectorAnd v0 v1)))

(define-cproc u8vector-and! (v0::<u8vector> v1)
  (return (Scm_U8VectorAndX v0 v1)))
(define-cproc u8vector-ior (v0::<u8vector> v1)
  (return (Scm_U8VectorIor v0 v1)))

(define-cproc u8vector-ior! (v0::<u8vector> v1)
  (return (Scm_U8VectorIorX v0 v1)))
(define-cproc u8vector-xor (v0::<u8vector> v1)
  (return (Scm_U8VectorXor v0 v1)))

(define-cproc u8vector-xor! (v0::<u8vector> v1)
  (return (Scm_U8VectorXorX v0 v1)))
(define-cproc s16vector-and (v0::<s16vector> v1)
  (return (Scm_S16VectorAnd v0 v1)))

(define-cproc s16vector-and! (v0::<s16vector> v1)
  (return (Scm_S16VectorAndX v0 v1)))
(define-cproc s16vector-ior (v0::<s16vector> v1)
  (return (Scm_S16VectorIor v0 v1)))

(define-cproc s16vector-ior! (v0::<s16vector> v1)
  (return (Scm_S16VectorIorX v0 v1)))
(define-cproc s16vector-xor (v0::<s16vector> v1)
  (return (Scm_S16VectorXor v0 v1)))

(define-cproc s16vector-xor! (v0::<s16vector> v1)
  (return (Scm_S16VectorXorX v0 v1)))
(define-cproc u16vector-and (v0::<u16vector> v1)
  (return (Scm_U16VectorAnd v0 v1)))

(define-cproc u16vector-and! (v0::<u16vector> v1)
  (return (Scm_U16VectorAndX v0 v1)))
(define-cproc u16vector-ior (v0::<u16vector> v1)
  (return (Scm_U16VectorIor v0 v1)))

(define-cproc u16vector-ior! (v0::<u16vector> v1)
  (return (Scm_U16VectorIorX v0 v1)))
(define-cproc u16vector-xor (v0::<u16vector> v1)
  (return (Scm_U16VectorXor v0 v1)))

(define-cproc u16vector-xor! (v0::<u16vector> v1)
  (return (Scm_U16VectorXorX v0 v1)))
(define-cproc s32vector-and (v0::<s32vector> v1)
  (return (Scm_S32VectorAnd v0 v1)))

(define-cproc s32vector-and! (v0::<s32vector> v1)
  (return (Scm_S32VectorAndX v0 v1)))
(define-cproc s32vector-ior (v0::<s32vector> v1)
  (return (Scm_S32VectorIor v0 v1)))

(define-cproc s32vector-ior! (v0::<s32vector> v1)
  (return (Scm_S32VectorIorX v0 v1)))
(define-cproc s32vector-xor (v0::<s32vector> v1)
  (return (Scm_S32VectorXor v0 v1)))

(define-cproc s32vector-xor! (v0::<s32vector> v1)
  (return (Scm_S32VectorXorX v0 v1)))
(define-cproc u32vector-and (v0::<u32vector> v1)
  (return (Scm_U32VectorAnd v0 v1)))

(define-cproc u32vector-and! (v0::<u32vector> v1)
  (return (Scm_U32VectorAndX v0 v1)))
(define-cproc u32vector-ior (v0::<u32vector> v1)
  (return (Scm_U32VectorIor v0 v1)))

(define-cproc u32vector-ior! (v0::<u32vector> v1)
  (return (Scm_U32VectorIorX v0 v1)))
(define-cproc u32vector-xor (v0::<u32vector> v1)
  (return (Scm_U32VectorXor v0 v1)))

(define-cproc u32vector-xor! (v0::<u32vector> v1)
  (return (Scm_U32VectorXorX v0 v1)))
(define-cproc s64vector-and (v0::<s64vector> v1)
  (return (Scm_S64VectorAnd v0 v1)))

(define-cproc s64vector-and! (v0::<s64vector> v1)
  (return (Scm_S64VectorAndX v0 v1)))
(define-cproc s64vector-ior (v0::<s64vector> v1)
  (return (Scm_S64VectorIor v0 v1)))

(define-cproc s64vector-ior! (v0::<s64vector> v1)
  (return (Scm_S64VectorIorX v0 v1)))
(define-cproc s64vector-xor (v0::<s64vector> v1)
  (return (Scm_S64VectorXor v0 v1)))

(define-cproc s64vector-xor! (v0::<s64vector> v1)
  (return (Scm_S64VectorXorX v0 v1)))
(define-cproc u64vector-and (v0::<u64vector> v1)
  (return (Scm_U64VectorAnd v0 v1)))

(define-cproc u64vector-and! (v0::<u64vector> v1)
  (return (Scm_U64VectorAndX v0 v1)))
(define-cproc u64vector-ior (v0::<u64vector> v1)
  (return (Scm_U64VectorIor v0 v1)))

(define-cproc u64vector-ior! (v0::<u64vector> v1)
  (return (Scm_U64VectorIorX v0 v1)))
(define-cproc u64vector-xor (v0::<u64vector> v1)
  (return (Scm_U64VectorXor v0 v1)))

(define-cproc u64vector-xor! (v0::<u64vector> v1)
  (return (Scm_U64VectorXorX v0 v1)))
(define-cproc s8vector-dot (v0::<s8vector> v1) Scm_VMS8VectorDotProd)
(define-cproc u8vector-dot (v0::<u8vector> v1) Scm_VMU8VectorDotProd)
(define-cproc s16vector-dot (v0::<s16vector> v1) Scm_VMS16VectorDotProd)
(define-cproc u16vector-dot (v0::<u16vector> v1) Scm_VMU16VectorDotProd)
(define-cproc s32vector-dot (v0::<s32vector> v1) Scm_VMS32VectorDotProd)
(define-cproc u32vector-dot (v0::<u32vector> v1) Scm_VMU32VectorDotProd)
(define-cproc s64vector-dot (v0::<s64vector> v1) Scm_VMS64VectorDotProd)
(define-cproc u64vector-dot (v0::<u64vector> v1) Scm_VMU64VectorDotProd)
(define-cproc f16vector-dot (v0::<f16vector> v1) Scm_VMF16VectorDotProd)
(define-cproc f32vector-dot (v0::<f32vector> v1) Scm_VMF32VectorDotProd)
(define-cproc f64vector-dot (v0::<f64vector> v1) Scm_VMF64VectorDotProd)
(define-cproc s8vector-range-check (v0::<s8vector> min max)
  Scm_S8VectorRangeCheck)
(define-cproc s8vector-clamp (v0::<s8vector> min max)
  Scm_S8VectorClamp)
(define-cproc s8vector-clamp! (v0::<s8vector> min max)
  Scm_S8VectorClampX)
(define-cproc u8vector-range-check (v0::<u8vector> min max)
  Scm_U8VectorRangeCheck)
(define-cproc u8vector-clamp (v0::<u8vector> min max)
  Scm_U8VectorClamp)
(define-cproc u8vector-clamp! (v0::<u8vector> min max)
  Scm_U8VectorClampX)
(define-cproc s16vector-range-check (v0::<s16vector> min max)
  Scm_S16VectorRangeCheck)
(define-cproc s16vector-clamp (v0::<s16vector> min max)
  Scm_S16VectorClamp)
(define-cproc s16vector-clamp! (v0::<s16vector> min max)
  Scm_S16VectorClampX)
(define-cproc u16vector-range-check (v0::<u16vector> min max)
  Scm_U16VectorRangeCheck)
(define-cproc u16vector-clamp (v0::<u16vector> min max)
  Scm_U16VectorClamp)
(define-cproc u16vector-clamp! (v0::<u16vector> min max)
  Scm_U16VectorClampX)
(define-cproc s32vector-range-check (v0::<s32vector> min max)
  Scm_S32VectorRangeCheck)
(define-cproc s32vector-clamp (v0::<s32vector> min max)
  Scm_S32VectorClamp)
(define-cproc s32vector-clamp! (v0::<s32vector> min max)
  Scm_S32VectorClampX)
(define-cproc u32vector-range-check (v0::<u32vector> min max)
  Scm_U32VectorRangeCheck)
(define-cproc u32vector-clamp (v0::<u32vector> min max)
  Scm_U32VectorClamp)
(define-cproc u32vector-clamp! (v0::<u32vector> min max)
  Scm_U32VectorClampX)
(define-cproc s64vector-range-check (v0::<s64vector> min max)
  Scm_S64VectorRangeCheck)
(define-cproc s64vector-clamp (v0::<s64vector> min max)
  Scm_S64VectorClamp)
(define-cproc s64vector-clamp! (v0::<s64vector> min max)
  Scm_S64VectorClampX)
(define-cproc u64vector-range-check (v0::<u64vector> min max)
  Scm_U64VectorRangeCheck)
(define-cproc u64vector-clamp (v0::<u64vector> min max)
  Scm_U64VectorClamp)
(define-cproc u64vector-clamp! (v0::<u64vector> min max)
  Scm_U64VectorClampX)
(define-cproc f16vector-range-check (v0::<f16vector> min max)
  Scm_F16VectorRangeCheck)
(define-cproc f16vector-clamp (v0::<f16vector> min max)
  Scm_F16VectorClamp)
(define-cproc f16vector-clamp! (v0::<f16vector> min max)
  Scm_F16VectorClampX)
(define-cproc f32vector-range-check (v0::<f32vector> min max)
  Scm_F32VectorRangeCheck)
(define-cproc f32vector-clamp (v0::<f32vector> min max)
  Scm_F32VectorClamp)
(define-cproc f32vector-clamp! (v0::<f32vector> min max)
  Scm_F32VectorClampX)
(define-cproc f64vector-range-check (v0::<f64vector> min max)
  Scm_F64VectorRangeCheck)
(define-cproc f64vector-clamp (v0::<f64vector> min max)
  Scm_F64VectorClamp)
(define-cproc f64vector-clamp! (v0::<f64vector> min max)
  Scm_F64VectorClampX)
(define-cproc s16vector-swap-bytes (v0::<s16vector>) Scm_S16VectorSwapBytes)
(define-cproc s16vector-swap-bytes!(v0::<s16vector>) Scm_S16VectorSwapBytesX)
(define-cproc u16vector-swap-bytes (v0::<u16vector>) Scm_U16VectorSwapBytes)
(define-cproc u16vector-swap-bytes!(v0::<u16vector>) Scm_U16VectorSwapBytesX)
(define-cproc s32vector-swap-bytes (v0::<s32vector>) Scm_S32VectorSwapBytes)
(define-cproc s32vector-swap-bytes!(v0::<s32vector>) Scm_S32VectorSwapBytesX)
(define-cproc u32vector-swap-bytes (v0::<u32vector>) Scm_U32VectorSwapBytes)
(define-cproc u32vector-swap-bytes!(v0::<u32vector>) Scm_U32VectorSwapBytesX)
(define-cproc s64vector-swap-bytes (v0::<s64vector>) Scm_S64VectorSwapBytes)
(define-cproc s64vector-swap-bytes!(v0::<s64vector>) Scm_S64VectorSwapBytesX)
(define-cproc u64vector-swap-bytes (v0::<u64vector>) Scm_U64VectorSwapBytes)
(define-cproc u64vector-swap-bytes!(v0::<u64vector>) Scm_U64VectorSwapBytesX)
(define-cproc f16vector-swap-bytes (v0::<f16vector>) Scm_F16VectorSwapBytes)
(define-cproc f16vector-swap-bytes!(v0::<f16vector>) Scm_F16VectorSwapBytesX)
(define-cproc f32vector-swap-bytes (v0::<f32vector>) Scm_F32VectorSwapBytes)
(define-cproc f32vector-swap-bytes!(v0::<f32vector>) Scm_F32VectorSwapBytesX)
(define-cproc f64vector-swap-bytes (v0::<f64vector>) Scm_F64VectorSwapBytes)
(define-cproc f64vector-swap-bytes!(v0::<f64vector>) Scm_F64VectorSwapBytesX)
